#!/usr/bin/env python
# coding: utf8
"""
Chainable pandas row-processing.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import re
import functools
import six
import numpy as np
import pandas as pd


def _get_func_argnames(func):
    # TODO: support defaults (kwargs)
    assert callable(func)
    func_code = func.__code__
    result = func_code.co_varnames[:func_code.co_argcount]
    if hasattr(func, '__self__'):
        # classmethods and bound methods, but not staticmethods.
        # XX: not certain about this.
        result = result[1:]
    return result


def _nmap_common(df, func, source=None, axis='columns', result_type='broadcast', **apply_kwargs):
    if source is None:
        source = _get_func_argnames(func)
    pre_df = df[list(source)]
    return pre_df.apply(
        # TODO: experiment with the performance of this part.
        lambda row: func(**row.to_dict()),
        axis=axis,
        # NOTE: requires pandas >= 0.23.0
        result_type=result_type,
        **apply_kwargs)


def nmap_one(df, func, target_column, source=None, inplace=False, **apply_kwargs):
    """
    Map df rows using `func(row)` and make a new column from the return value.
    """
    results = _nmap_common(df, func, source=source, result_type='reduce')
    if not inplace:
        df = df.copy()
    df[target_column] = results
    return df


def nmap(df, func, target=None, source=None, inplace=False, **apply_kwargs):
    """
    Map df rows using `func(row)` and make multiple new columns from the iterable return value.

    ...

    :param target: names of the target columns.
    """
    results = _nmap_common(df, func, source=source, result_type='expand')

    if not target:
        if isinstance(results.columns, pd.RangeIndex):
            raise ValueError("Refusing to put in new columns from a RangeIndex; specify the `target` or return dicts")
        target = results.columns
    else:
        if len(results.columns) != len(target):
            raise ValueError("`target` length mismatch: result has %r, target has %r" % (
                len(results.columns), len(target)))

    if not inplace:
        df = df.copy()

    for src, dst in zip(results.columns, target):
        df[dst] = results[src]

    return df


_nfilter_extras = dict(
    # name -> func(series, value): bool_series
    {'in': lambda series, value: series.map(lambda cell: value in cell)},
    isnull=lambda series, value: pd.isnull(series) if value else ~pd.isnull(series),  # pylint: disable=invalid-unary-operand-type
    defined=lambda series, value: ~pd.isnull(series) if value else pd.isnull(series),  # pylint: disable=invalid-unary-operand-type
    nonzero=lambda series, value: series.map(bool) if value else ~series.map(bool),
    zero=lambda series, value: ~series.map(bool) if value else series.map(bool),
    re=lambda series, value, **kwargs: series.map(lambda cell: bool(re.search(value, cell, **kwargs))),
    startswith=lambda series, value: series.map(lambda cell: cell.startswith(value)),
    endswith=lambda series, value: series.map(lambda cell: cell.endswith(value)),
    gt=lambda series, value: series.map(lambda cell: cell > value),
    gte=lambda series, value: series.map(lambda cell: cell >= value),
    lt=lambda series, value: series.map(lambda cell: cell < value),
    lte=lambda series, value: series.map(lambda cell: cell <= value),
    # ...
)


def _nfilter_one_by_extras(df, name, value):
    parts = name.split("__", 1)
    if len(parts) < 2:
        raise ValueError("Incorrect usage of _nfilter_one_by_extras")
    name, extra = parts

    flt_func = _nfilter_extras.get(extra)
    if flt_func is None:
        raise ValueError("Unknown extra-filter", extra)

    return flt_func(df[name], value)


def nfilter_base_gen(df, *args, **kwargs):
    """
    DRY concise filters.

    Supported filters:

      * `lambda field1, field2: ...`
      * `name__defined=True`
      * `name__in=[...]`
      * `name__re="..."` (contains)
      * `name__{modifier}`
      * `name=lambda value: ...`  # uncertain, subject to change
      * `name=value`
    """
    for arg in args:
        flt = _nmap_common(df, func=arg, result_type='reduce')
        if flt.dtype.name != 'bool':
            flt = flt.map(bool)  # avoid some inconveniences
        yield flt

    for name, value in kwargs.items():
        flt = None
        if "__" in name:
            yield _nfilter_one_by_extras(df, name, value)
        elif callable(value):
            target = _nmap_common(df, func=value, source=[name], result_type='reduce')
            yield df[name] == target
        else:
            yield df[name] == value


def nfilter_base(df, *args, **kwargs):
    return list(nfilter_base_gen(df, *args, **kwargs))


def _nfilter_to_selection(df, *args, **kwargs):
    filters = nfilter_base(df, *args, **kwargs)
    if not filters:
        return df
    selected = functools.reduce(lambda left, right: left & right, filters)
    return selected


def nfilter(df, *args, **kwargs):
    """
    Shorthand for 'where X and Y and ...'.
    """
    selected = _nfilter_to_selection(df, *args, **kwargs)
    return df[selected]


# TODO?: nfilter_any (for `or` over the arguments)
# Probably less useful than writing an explicit `nfilter(lambda col1, col2: col1 or col2)`

def nexclude(df, *args, **kwargs):
    """
    Shorthand for 'where not (X and Y and ...)'
    """
    selected = _nfilter_to_selection(df, *args, **kwargs)
    return df[~ selected]


def nproject_one(df, value, **kwargs):
    if callable(value):
        return _nmap_common(df, func=value, result_type='reduce')

    if isinstance(value, six.string_types):
        return df[value]  # simple rename

    raise Exception("Unexpected nproject_one value", value)


def nproject(df, *args, **kwargs):
    """
    :param args: Either of:
      * ``: only set new columns as a result of processing.
      * `'-'`: do not return the original columns.
      * `'-col1', '-col2'`: exclude the specified columns from the source.
      * `'col1', 'col2'`: leave only the specified columns from the source.

    :param kwargs:
      * column_name -> column_expression
        * column_name: name of the target (output) column.
        * column expression: describes how the value will be obtained. Either of:
          * string: will make a copy of another column (useful for renaming).
          * `lambda col1, col2, ...: result_value` - will make a new column from the call result.
    """

    # # TODO?: _inplace
    # :param _inplace: (default: False) whether the passed dataframe should be mutated.
    # inplace = kwargs.pop('_inplace', False)

    if not args:
        # leave the columns as-is
        res_columns = list(df.columns)
    elif any(isinstance(val, six.string_types) and val.startswith("-") for val in args):
        # Exclude some columns
        assert all(val.startswith("-") for val in args), "should either be all inclusive or all exclusive"
        if len(args) == 1 and args[0] == '-':
            res_columns = []
        else:
            exclude_fields = set(val[1:] for val in args)
            # unknown_excluded = exclude_fields - set(df.columns)
            res_columns = list(col for col in df.columns if col not in exclude_fields)
    else:
        res_columns = args

    result = df[res_columns]  # supposed to make a shallow copy too

    # TODO?: support automatic ordering of the projections for new-column dependencies.
    # e.g. `a=lambda col1: ..., b=lambda a: ...`

    for key, value in kwargs.items():
        res_value = nproject_one(df, value, key=key)
        if res_value is not None:
            result[key] = res_value

    return result


def nsort(df, *args, **sort_values_kwargs):
    """
    `df.sort_values` that allows temporary fields, in form of callables (same
    as `nproject`).

    :param args: fields / callables, ordered, to be used for sorting.

    :param sort_values_kwargs: additional arguments to `pandas.DataFrame.sort_values`

    Example:

        df.nsort(
            'a',
            lambda b: 1 if b.startswith('x') else 2,
            'b',
            ascending=[True, False, False],
            na_position='first')
    """
    order_columns = ['x{:02d}'.format(idx) for idx in range(len(args))]
    projection = dict(zip(order_columns, args))
    order_df = nproject(df, '-', **projection)
    order_df.sort_values(order_columns, **sort_values_kwargs)
    return df.loc(order_df.index)


def patch_all():
    pd.DataFrame.nmap_one = nmap_one
    pd.DataFrame.nmap = nmap
    pd.DataFrame.nfilter = nfilter
    pd.DataFrame.nexclude = nexclude
    pd.DataFrame.nproject = nproject
    pd.DataFrame.nsort = nsort


def _examples():
    """ ... """

    # WARNING:
    patch_all()

    df = pd.DataFrame(dict(
        col1=['A', 'A', 'B', np.nan, 'D', 'C'],
        col2=[2, 1, 9, 8, 7, 4],
        col3=[0, 1, 9, 4, 2, 3],
    ))
    # make a copy of the column
    dfp = df.nproject(col3cp='col3')
    assert list(dfp.columns) == list(df.columns) + ['col3cp'], dfp.columns
    # process a couple columns into a new one
    dfp = df.nproject(col23=lambda col2, col3: col2 + col3 * 2)
    assert list(dfp.columns) == list(df.columns) + ['col23'], dfp.columns
    # Reminder: the simple stuff can be done in this more performant way; the
    # `nproject` method is for more tricky per-row computations.
    assert (dfp['col23'] == (dfp['col2'] + dfp['col3'] * 2)).all(), dfp['col23']

    # process a couple columns into a couple new columns
    dfp = df.nmap(lambda col2, col3: (col2 * col3, col3 / col2), target=('colm', 'cold'))
    assert list(dfp.columns) == list(df.columns) + ['colm', 'cold'], dfp.columns

    # Same but returning a dict
    dfp = df.nmap(lambda col2, col3: dict(colm=col2 * col3, cold=col3 / col2))
    assert (
        (list(dfp.columns) == list(df.columns) + ['colm', 'cold']) or
        (list(dfp.columns) == list(df.columns) + ['cold', 'colm'])), dfp.columns

    dfp = df.nfilter(col2__gte=3)
    assert list(dfp['col3']) == [9, 4, 2, 3], dfp['col3']


if __name__ == '__main__':
    _examples()



# coding: utf8
"""
WARNING: a monkeypatch-on-import wrapper of `pd_streamlib`.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

from . import pd_streamlib
pd_streamlib.patch_all()



# coding: utf8
""" A reimplementation of ZeroMQ logging handler. Based on
python-zeromq-log.

General idea:
* ZMQHandler is initialized with zmq socket (or with an uri using
 ZMQHandler.to) and added as a handler to some logger.
* When a record is emitted, it is pickled (NOTE: user-set formatter is
 not supposed to be used) and sent over the socket, as a multipart
 message, together with the `channel`.
* Logging server simply unpickles and handles the record with its
 configured logging.

Required Twisted and TxZMQ.

Note: there is also zmq.log.handlers.PUBHandler, but it converts the
whole message to string before sending rather than doing the final
combining at the receiver's side.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings
# import traceback
try:
    import cPickle as pickle
except ImportError:
    import pickle
import logging
import zmq
# import zmq.log.handlers as zmqlog
import txzmq
from twisted.internet import reactor


class ZMQFormatter(logging.Formatter):
    _pickle_protocol = -1  # NOTE.
    def makePickle(self, record):
        """ Pickle the record. Code copy from
        logging.handlers.SocketHandler. Does not prepend size.  Possibly
        modifies the record. """
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text ...
            #dummy = self.format(record)  # infiniterecursion!
            dummy = logging.Formatter.format(self, record)
            record.exc_info = None  # to avoid Unpickleable error
        # See issue #14436: If msg or args are objects, they may not be
        # available on the receiving end. So we convert the msg % args
        # to a string, save it as msg and zap the args.
        d = dict(record.__dict__)  # NOTE: might be preferable to remove this later.
        d['msg'] = record.getMessage()
        d['args'] = None
        s = pickle.dumps(d, self._pickle_protocol)
        if ei:
            record.exc_info = ei  # for next handler
        #slen = struct.pack(">L", len(s))
        #return slen + s
        return s
    def format(self, record):
        """ Format: pickles the record """
        #data = record.__dict__.copy()  # will be preferable to avoid copying
        #data = record
        #if data.get('traceback'):
        #    data['traceback'] = self.formatException(data['traceback'])
        #data['time'] = data['time'].isoformat()  # hopefully unnecessary
        #return pickle.dumps(data)
        return self.makePickle(record)
class ZMQHandler(logging.Handler):
    @classmethod
    def to(cls, uri, *ar, **kwa):
        context = zmq.Context.instance()
        #context = zmq.Context()  # separate thread, maybe
        publisher = context.socket(zmq.PUSH)
        # NOTE: do not wait indefinitely for message sending. Might be a
        #  bad idea though.
        publisher.setsockopt(zmq.LINGER, 7000)
        #publisher.setsockopt(zmq.SNDHWM, 500)
        # ...
        publisher.connect(uri)
        return cls(publisher, *ar, **kwa)
    def __init__(self, zmq_client, name='', channel='log',
                 level=logging.NOTSET):
        super(ZMQHandler, self).__init__(level)
        self.zmq_client = zmq_client
        self.channel = channel
        self.formatter = ZMQFormatter()
    def emit(self, record):
        try:
            msg = self.format(record)
            self.send(msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)
    def setFormatter(self, fmt):
        warnings.warn("ZMQHandler's formatter is not supposed to be set.")
    def send(self, data):
        #try:
        self.zmq_client.send_multipart([self.channel, data], zmq.NOBLOCK)
        ## Handler by handleError anyway.
        #except zmq.ZMQError as e:
        #    print("ZMQHandler error")
        #    traceback.print_stack()


class TxLogServer(object):
    """ TxZMQ-based LogServer for the ZMQHandler.  Uses only PULL_BIND
    socket; copy and modify for other cases. """
    def __init__(self, uri, factory=None, run=False):
        self._socket = self._init_zmq(uri, factory=factory)
        self.log_protocol = LogProtocol()
        if run:
            self.run()
    def run(self, verbose=True):
        if verbose:
            print("TxLogServer: Starting.")
        return reactor.run()  # pylint: disable=E1101

    def _init_zmq(self, uri, factory=None):
        socket = txzmq.ZmqPullConnection(
            factory or txzmq.ZmqFactory(),
            txzmq.ZmqEndpoint('bind', uri))
        socket.onPull = self.on_message
        return socket
    def on_message(self, message):
        ## Exception handling is not used here as this is normally
        ##   called from Twisted.
        ## Expecting a multipart message with (topic, recordpickle)
        self.log_protocol.handle_msg(message[0], message[1])
class TxLogServerSub(TxLogServer):
    """ SUB_BIND version """
    def _init_zmq(self, uri, factory=None):
        _zf = factory or txzmq.ZmqFactory()
        _ze = txzmq.ZmqEndpoint('bind', uri)
        socket = txzmq.ZmqSubConnection(_zf, _ze)
        socket.subscribe('')
        socket.gotMessage = self.on_message
        return socket
    def on_message(self, message, tag):
        self.log_protocol.handle_msg(tag, message)
class LogProtocol(object):
    """ Actual data handler for the LogServer """
    def __init__(self):
        self._log = logging.getLogger("logsrv")
        self._log.setLevel(0)
    def handle_msg(self, channel, record_pickle):
        record_data = pickle.loads(record_pickle)
        self.handle(channel, record_data)
    def handle(self, channel, record_data):
        record = logging.makeLogRecord(record_data)
        ## handle `channel`:
        record.name = '{%s} %s' % (channel, record.name)
        # print(record.__dict__)  # dbg
        self._log.handle(record)

# # Sample record data:
# # Normal:
# {'relativeCreated': 925.4779815673828, 'process': 901, 'module': 'tst_zmqlog2', 'funcName': '<module>', 'filename': 'tst_zmqlog2.py', 'levelno': 30, 'processName': 'MainProcess', 'lineno': 12, 'msg': "tst, warn: <open file '/dev/null', mode 'r' at 0x2c5de40>", 'args': None, 'exc_text': None, 'name': 'wut', 'thread': 140164311705344, 'created': 1359000458.195124, 'threadName': 'MainThread', 'msecs': 195.12391090393066, 'pathname': './tst_zmqlog2.py', 'exc_info': None, 'levelname': 'WARNING'}
# # Traceback:
# {'relativeCreated': 925.713062286377, 'process': 901, 'module': 'tst_zmqlog2', 'funcName': '<module>', 'message': 'more tst', 'filename': 'tst_zmqlog2.py', 'levelno': 40, 'processName': 'MainProcess', 'lineno': 18, 'msg': 'more tst', 'args': None, 'exc_text': 'Traceback (most recent call last):\n  File "./tst_zmqlog2.py", line 16, in <module>\n    b = 1/0\nZeroDivisionError: integer division or modulo by zero', 'name': 'wut', 'thread': 140164311705344, 'created': 1359000458.195359, 'threadName': 'MainThread', 'msecs': 195.3589916229248, 'pathname': './tst_zmqlog2.py', 'exc_info': None, 'levelname': 'ERROR'}


def test1():
    import pyaux
    pyaux.runlib.init_logging()
    root_logger = logging.getLogger()
    zh = ZMQHandler.to('ipc://var/log.ipc', channel='the_name')
    root_logger.addHandler(zh)
    root_logger.warn("tst: %r", open('/dev/null'))

    try:
        b = 1/0
    except Exception as e:
        root_logger.exception("more tst")


def test2():
    ## NOTE: Suggested way for configuring the logging is django-style
    ##   settings module with `init_logging(process_name)` function.
    #import settings
    #settings.init_logging("tst_log2")
    import pyaux
    pyaux.runlib.init_logging()
    ## ...
    log = logging.getLogger("wut")

    #time.sleep(0.5)  # connect race condition :(
    #log.debug("tst, dbg: %r", open('/dev/null'))
    #log.info("tst, info: %r", open('/dev/null'))
    log.warn("tst, warn: %r", open('/dev/null'))
    #log.error("tst, err: %r", open('/dev/null'))

    try:
        b = 1/0
    except Exception as e:
        log.exception("more tst")

    print("DONE.")


if __name__ == '__main__':
    import sys
    target = 'test1'
    if len(sys.argv) > 1:
        target = sys.argv[-1]
    ## ...
    if target == 'test1':
        test1()
    elif target == 'test2':
        test2()



#!/usr/bin/env python

from __future__ import print_function, unicode_literals, absolute_import, division

import os
from setuptools import setup, find_packages

__version__ = "1.2.0"

DESCRIPTION = "Collection of helpers and useful things for Python: library-specific pieces"

LONG_DESCRIPTION = """
"""

try:
    LONG_DESCRIPTION = (
        LONG_DESCRIPTION +
        open(os.path.join(os.path.dirname(__file__), "README.rst"), "rb").read().decode("utf-8"))
except Exception as _exc:
    print("Pkg-description error:", _exc)


SETUP_KWARGS = dict(
    name="pyauxm",
    version=__version__,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    # classifiers=[],
    # keywords="...,...",
    author="HoverHell",
    author_email="hoverhell@gmail.com",
    url="https://github.com/HoverHell/pyauxm",
    download_url="https://github.com/HoverHell/pyauxm/tarball/%s" % (__version__,),
    packages=find_packages(),
    install_requires=["six"],
    extras_require={
        # All things that are known to be used in some part of this
        # library or another.
        "known": [
            "django",  # in the psql helper
            "Twisted",  # bunch of twisted stuff here
            "Cython",  # at least one pyx module
            "pandas",  # here and there
            "Pygments",  # json / yaml coloring
            "pylzma",  # helpers for it
            "simplejson",  # optional, for speed
            "pp",
            "line_profiler",
            "pyzmq",
        ],
    },
)


if __name__ == "__main__":
    setup(**SETUP_KWARGS)
